@@ -191,7 +191,7 @@ istream &operator>>(istream &s, qd_real &qd) {
 ostream &operator<<(ostream &os, const qd_real &qd) {
   bool showpos = (os.flags() & ios_base::showpos) != 0;
   bool uppercase = (os.flags() & ios_base::uppercase) != 0;
-  return os << qd.to_string(os.precision(), os.width(), os.flags(), 
+  return os << qd.to_string(os.precision(), os.width(), os.flags(),
       showpos, uppercase, os.fill());
 }
 
@@ -346,9 +346,9 @@ void qd_real::to_digits(char *s, int &expn, int precision) const {
   }
 
   /* If first digit is 10, shift everything. */
-  if (s[0] > '9') { 
-    e++; 
-    for (i = precision; i >= 2; i--) s[i] = s[i-1]; 
+  if (s[0] > '9') {
+    e++;
+    for (i = precision; i >= 2; i--) s[i] = s[i-1];
     s[0] = '1';
     s[1] = '0';
   }
@@ -369,6 +369,40 @@ void qd_real::write(char *s, int len, int precision,
   s[len-1] = 0;
 }
 
+void round_string_qd(char *s, int precision, int *offset){
+	/*
+	 Input string must be all digits or errors will occur.
+	 */
+
+	int i;
+	int D = precision ;
+
+	/* Round, handle carry */
+	  if (D>0 && s[D] >= '5') {
+	    s[D-1]++;
+
+	    i = D-1;
+	    while (i > 0 && s[i] > '9') {
+	      s[i] -= 10;
+	      s[--i]++;
+	    }
+	  }
+
+	  /* If first digit is 10, shift everything. */
+	  if (s[0] > '9') {
+	    // e++; // don't modify exponent here
+	    for (i = precision; i >= 1; i--) s[i+1] = s[i];
+	    s[0] = '1';
+	    s[1] = '0';
+
+	    (*offset)++ ; // now offset needs to be increased by one
+	    precision++ ;
+	  }
+
+	  s[precision] = 0; // add terminator for array
+}
+
+
 string qd_real::to_string(int precision, int width, ios_base::fmtflags fmt, 
     bool showpos, bool uppercase, char fill) const {
   string s;
@@ -407,19 +441,50 @@ string qd_real::to_string(int precision, int width, ios_base::fmtflags fmt,
       int off = (fixed ? (1 + to_int(floor(log10(abs(*this))))) : 1);
       int d = precision + off;
 
+      int d_with_extra = d;
+      if(fixed)
+    	  d_with_extra = std::max(120, d); // longer than the max accuracy for DD
+
+      // highly special case - fixed mode, precision is zero, abs(*this) < 1.0
+      // without this trap a number like 0.9 printed fixed with 0 precision prints as 0
+      // should be rounded to 1.
+      if(fixed && (precision == 0) && (abs(*this) < 1.0)){
+    	  if(abs(*this) >= 0.5)
+    		  s += '1';
+    	  else
+    		  s += '0';
+
+    	  return s;
+      }
+
+      // handle near zero to working precision (but not exactly zero)
       if (fixed && d <= 0) {
         s += '0';
         if (precision > 0) {
           s += '.';
           s.append(precision, '0');
         }
-      } else {
-        char *t = new char[d+1];
+      } else {  // default
+
+        char *t ; // = new char[d+1];
         int j;
 
-        to_digits(t, e, d);
+        if(fixed){
+        	t = new char[d_with_extra+1];
+        	to_digits(t, e, d_with_extra);
+        }
+        else{
+        	t = new char[d+1];
+        	to_digits(t, e, d);
+        }
+
+        off = e + 1;
 
         if (fixed) {
+          // fix the string if it's been computed incorrectly
+          // round here in the decimal string if required
+          round_string_qd(t, d, &off);
+
           if (off > 0) {
             for (i = 0; i < off; i++) s += t[i];
             if (precision > 0) {
@@ -443,6 +508,36 @@ string qd_real::to_string(int precision, int width, ios_base::fmtflags fmt,
       }
     }
 
+    // trap for improper offset with large values
+    // without this trap, output of values of the for 10^j - 1 fail for j > 28
+    // and are output with the point in the wrong place, leading to a dramatically off value
+    if(fixed && (precision > 0)){
+    	// make sure that the value isn't dramatically larger
+    	double from_string = atof(s.c_str());
+
+    	// if this ratio is large, then we've got problems
+    	if( fabs( from_string / this->x[0] ) > 3.0 ){
+
+    		int point_position;
+
+    		// loop on the string, find the point, move it up one
+    		// don't act on the first character
+    		for(i=1; i < s.length(); i++){
+    			if(s[i] == '.'){
+    				s[i] = s[i-1] ;
+    				s[i-1] = '.' ;
+    				break;
+    			}
+    		}
+
+        	from_string = atof(s.c_str());
+        	// if this ratio is large, then the string has not been fixed
+        	if( fabs( from_string / this->x[0] ) > 3.0 ){
+        		dd_real::error("Re-rounding unsuccessful in large number fixed point trap.") ;
+        	}
+    	}
+    }
+
     if (!fixed) {
       /* Fill in exponent part */
       s += uppercase ? 'E' : 'e';
@@ -640,37 +735,53 @@ qd_real qd_real::accurate_div(const qd_real &a, const qd_real &b) {
   return qd_real(q0, q1, q2, q3);
 }
 
-QD_API qd_real sqrt(const qd_real &a) {
-  /* Strategy:  
+QD_API qd_real fsqrt(const qd_real &a, int &flag) {
+  /* Uses Heron's method, see:
+     https://en.wikipedia.org/wiki/Methods_of_computing_square_roots#Babylonian_method
 
-     Perform the following Newton iteration:
-
-       x' = x + (1 - a * x^2) * x / 2;
-       
-     which converges to 1/sqrt(a), starting with the
-     double precision approximation to 1/sqrt(a).
-     Since Newton's iteration more or less doubles the
-     number of correct digits, we only need to perform it 
-     twice.
+     1. x0 = approximate sqrt(a);
+     2. x_{n+1} = (1/2) * (x_n + a / x_n);
+     3. repeat 2 until corrections are small
   */
 
+  int i;
+  double e, eps;
+
+  qd_real r, diff;
+  qd_real half = "0.5000000000000000000000000000000000"
+                 "000000000000000000000000000000000000";
+
   if (a.is_zero())
-    return 0.0;
+    return (qd_real) 0.0;
 
   if (a.is_negative()) {
     qd_real::error("(qd_real::sqrt): Negative argument.");
     return qd_real::_nan;
   }
 
-  qd_real r = (1.0 / std::sqrt(a[0]));
-  qd_real h = mul_pwr2(a, 0.5);
+  eps = std::numeric_limits<qd_real>::epsilon();
 
-  r += ((0.5 - h * sqr(r)) * r);
-  r += ((0.5 - h * sqr(r)) * r);
-  r += ((0.5 - h * sqr(r)) * r);
+  qd_real x = std::sqrt(a[0]);
+  qd_real y;
 
-  r *= a;
-  return r;
+  for (i=0; i < 10; i++) {
+      y = half * (x + a / x);
+      diff = x - y;
+      x = y;
+      e = fabs(((diff[3] + diff[2]) + diff[1]) + diff[0]);
+      if (e < fabs(x.x[0]) * eps) {
+          flag = 0; // convergence achieved
+          return x;
+      }
+  }
+
+  flag = 1; // failed to converge
+  return x;
+}
+
+QD_API qd_real sqrt(const qd_real &a) {
+  int flag;
+  return fsqrt(a, flag);
 }
 
 
@@ -688,19 +799,40 @@ qd_real nroot(const qd_real &a, int n) {
      we only need to perform it twice.
 
    */
+	if (n <= 0) {
+		qd_real::error("(qd_real::nroot): N must be positive.");
+		return qd_real::_nan;
+	}
 
-  if (a == 0.0) {
-    return qd_real(0.0);
-  }
+	if (n % 2 == 0 && a.is_negative()) {
+		qd_real::error("(qd_real::nroot): Negative argument.");
+		return qd_real::_nan;
+	}
 
-  qd_real r = std::pow(a[0], -1.0/n);
+	if (n == 1) {
+		return a;
+	}
+	if (n == 2) {
+		return sqrt(a);
+	}
+	if (a.is_zero()) {
+		return qd_real(0.0);
+	}
 
-  double dbl_n = static_cast<double>(n);
-  r += r * (1.0 - a * (r ^ n)) / dbl_n;
-  r += r * (1.0 - a * (r ^ n)) / dbl_n;
-  r += r * (1.0 - a * (r ^ n)) / dbl_n;
 
-  return 1.0 / r;
+	/* Note  a^{-1/n} = exp(-log(a)/n) */
+	qd_real r = abs(a);
+	qd_real x = std::exp(-std::log(r.x[0]) / n);
+
+	/* Perform Newton's iteration. */
+	double dbl_n = static_cast<double>(n);
+	x += x * (1.0 - r * npwr(x, n)) / dbl_n;
+	x += x * (1.0 - r * npwr(x, n)) / dbl_n;
+	x += x * (1.0 - r * npwr(x, n)) / dbl_n;
+	if (a[0] < 0.0){
+		x = -x;
+	}
+	return 1.0 / x;
 }
 
 static const int n_inv_fact = 15;
